// RTPAgent.java
// Last Modified on May 30, 2000
// Final Version

import java.io.*;
import java.net.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;

import javax.swing.*;

import javax.media.*;
import javax.media.rtp.*;
import javax.media.format.*;
import javax.media.control.*;
import javax.media.protocol.*;
import javax.media.rtp.rtcp.*;
import javax.media.rtp.event.*;

public class RTPAgent extends JFrame
implements ControllerListener, ReceiveStreamListener, RemoteListener
{
  // localSessionAddress = Address from which this JMF receives audio 
  // remoteSessionAddress = Address to which this JMF sends audio

  private String localSessionAddress;
  private int localSessionPort;
  private static final int ttl = 1;

  // destination
  // rtp://remoteSessionAddress:remoteSessionPort/audio
  private String remoteSessionAddress;
  private int remoteSessionPort;

  private Processor processor = null;
 
  private SessionManager sendingMgr;    // SessionManager to send RTP data 
  private SessionManager receivingMgr;  // SessionManager to receive RTP data
  private SessionAddress remoteSession; // destination 
  private SessionAddress localSession;

  private DataSource processorOutput = null;	
  private SendStream streamSend = null; // sending stream within the session

  private Vector playerList = new Vector();
  
  private boolean started = false; // indication that RTP Agent has started

  private Vector participantList = new Vector(); // Session participants
  private JPanel panel = null;
  private JTextArea textArea = null;
  
  // constructor
  public RTPAgent() 
  {
    getLocalIPPort();
    start();
  }

    
  //************* Interface between JMF and MGCP ********************

  // Return the local address of the JMF Agent
  public String getLocalSessionAddress() 
  {
    return localSessionAddress;
  }

  // Return the local port number of the JMF Agent
  public int getLocalSessionPort() 
  {
    return localSessionPort;
  }

  // Used for both INVITE and REINVITE
  // Display the GUI for the source description and start transmitting
  //	to the destination (remote session) 
  public void setRemoteSession(String remoteAddress, int remotePort) 
  {
    remoteSessionAddress = remoteAddress;
    remoteSessionPort = remotePort;

    // When first called, transmit() is called to transmit to the 
    //	destination (INVITE)
    // When called again, handoff() is called to stop transmitting to the
    //	old destination and start sending to the new destination (REINVITE) 
    if (!started)
      transmit();
    else
      handoff();
  }

  // Used for BYE
  // Return true when successfully stopping transmitting to the destination
  public boolean stopTransmission() 
  {
    if (sendingMgr != null) {
      sendingMgr.closeSession("RTP Session Terminated");
      sendingMgr = null;
      return true;
    }
    else 
      return false;
  }
  
  //******************************************************************
 

  // Obtains the local address and local port for this JMF agent
  private void getLocalIPPort() 
  {
    // Obtain the local address
    try {
      localSessionAddress = InetAddress.getLocalHost().getHostAddress();
    }
    catch(UnknownHostException e) {
      System.err.println(e);
      System.exit(-1);
    }

    // Obtain an available local port
    LocalPort localPort = new LocalPort();
    localSessionPort = localPort.getLocalPort();
	
    if (localSessionPort == -1) {
      System.err.println("Can't get local port");
      System.exit(-1);
    }
	
  }

  // Get ready to capture audio through audio capture device, 
  //	to generate RTP-encoded data and to receive RTP packets
  private void start() 
  {
    // create a Session Manager to receive RTP packets sent to the 
    //	local session address of this JMF
    receivingMgr = createManager(localSessionAddress, localSessionPort,
				 false);
    if (receivingMgr == null) 
      System.err.println("Failed to create SessionManager");
    receivingMgr.addReceiveStreamListener(this);
    receivingMgr.addRemoteListener(this);
	
    // Create Processor to process on the captured audio data
    createProcessor();

    // Set the output content descriptor
    setOutputCD();
    
    //Program track to output RTP-encoded data
    programTrack();
    
    // Retrieve processor output
    getProcessorOutput();
    
  }

  // Create a processor for the specified media locator
  // and program it to output 
  private void createProcessor() {

    // DataSource that captures live audio
    // LINEAR, 44100.0 Hz, 16-bit, Mono,  BigEndian, Signed
    AudioFormat audio_format = 
      new AudioFormat(AudioFormat.LINEAR, 44100, 16, 1);
    Vector deviceList = CaptureDeviceManager.getDeviceList(audio_format);

    CaptureDeviceInfo deviceInfo = null;
    if (deviceList.size() > 0) {
      deviceInfo = (CaptureDeviceInfo)deviceList.firstElement();
    }
    else {
      // exit if we could not find the relevant capture device
      handleError("No appropriate capture device found");
    }
    // Create a processor for this capture device & exit if we
    // 	cannot create it
    System.out.println("Capture Device Name = " + deviceInfo.getName());
    System.out.println("Capture Device Locator = " + 
		       deviceInfo.getLocator().toString());

    try {
      processor = Manager.createProcessor(deviceInfo.getLocator());
    } catch (NoProcessorException npe) {
      handleError("Failed to create Processor");
    } catch (IOException ioe) {
      handleError("IOException while creating Processor");
    } 

    processor.addControllerListener(this);
  }
  
  // Set the output content descriptor
  private void setOutputCD() {
    
    // Put the Processor into Configured state
    processor.configure();

    // Wait for it to configure
    boolean result = handleStateTransition(processor, Processor.Configured);
    if (result == false)
      handleError("Couldn't configure processor");

    // Set the output content descriptor 
    processor.
      setContentDescriptor(new ContentDescriptor(
						 ContentDescriptor.RAW));
  }
  
  // Program track to output RTP-encoded data 
  private void programTrack() 
  {
    // Get a TrackControl for each track in the media stream
    TrackControl [] trackList = processor.getTrackControls();
    if (trackList == null || trackList.length < 1)
      handleError("No TrackControl available for this Processor");
			
    boolean programmed = false;
        
    // RTP Format: G.711(U-law), 8 kHz, 8-bit 
    AudioFormat rtpFormat = 
      new AudioFormat(AudioFormat.ULAW_RTP, 8000, 8, 1);
    
    // Program one of the tracks to output ULAW_RTP data
    for (int i = 0; i < trackList.length; i++) {
      if (!programmed && trackList[i] instanceof FormatControl) {
	// Set the data format
	Format control = ((FormatControl)
	  trackList[i]).setFormat(rtpFormat);      
	
	if (control == null) {
	  trackList[i].setEnabled(false);
	}
	else
	  programmed = true;
      } 
      else
	// Disable the track that can't be set to ULAW
	trackList[i].setEnabled(false);
    }

    if (!programmed)
      handleError("No audio track found");
  }
  
  // Retrieve Processor output
  private void getProcessorOutput() 
  {
    // Realize the processor 
    processor.realize();
	
    // Wait until Processor gets realized. 
    boolean result = handleStateTransition(processor, Processor.Realized);
    if (result == false)
      handleError("Failed to realize processor");

    // Retrieve the output of the Processor as a DataSource 
    // and exit if we fail  
    try {  
      processorOutput = processor.getDataOutput(); 
    } catch (NotRealizedError e) {
      handleError("Failed to retrieve the processor output");
    }
    if (processorOutput == null) 
      handleError("Processor Output is NULL");
    if (!(processorOutput instanceof PushBufferDataSource)) 
      handleError("Processor Output is not a PushBufferDataSource");
	
    try {
      processorOutput.connect();
      processorOutput.start();
    }
    catch (IOException e) {
      handleError("IOException in Processor Ouput");
    }
  }

  // Destroy Processor due to error
  public void destroyProcessor() {
    synchronized (this) {
      if (processor != null) {
	processor.stop();
	processor.close();
	processor = null;
      }
    }
  }

  // Handle Processor error
  private void handleError(String errorMessage) {
    System.err.println(errorMessage);
    // deallocate resources
    destroyProcessor();
    closeManager();
    System.exit(1);
  }
  

  // Create RTP Session Manager
  private SessionManager createManager(String sessionAddress, 
				       int sessionPort,
				       boolean isSender) 
  {
    SessionManager newMgr = 
      (SessionManager)new com.sun.media.rtp.RTPSessionMgr();
    if (newMgr == null)
      return null;

    // buffer control to reduce audio delay
    Control c = 
      (Control)newMgr.getControl("javax.media.control.BufferControl");
    if (c != null) {
      ((BufferControl)c).setBufferLength(16);
      System.out.println("Buffer length set to " + 
			 ((BufferControl)c).getBufferLength());
    }

    if (isSender) {
      InetAddress remoteSessionAddr = null;
      try {
	remoteSessionAddr = InetAddress.getByName(remoteSessionAddress);
      } catch (UnknownHostException e){
	System.err.println("inetaddress " + e.getMessage());
	e.printStackTrace();
	return null;
      }	    

      // destination
      remoteSession = 
	new SessionAddress(remoteSessionAddr,
			   remoteSessionPort,
			   remoteSessionAddr,
			   remoteSessionPort + 1);
      System.out.println("Remote Session Address = " + 
		       remoteSessionAddr);
      System.out.println("Remote Session port = " + 
			 remoteSessionPort);
    }
    else {
      InetAddress localSessionAddr = null;
      try {
	localSessionAddr = InetAddress.getByName(sessionAddress);
      }
      catch (UnknownHostException e){
	System.err.println("inetaddress " + e.getMessage());
	e.printStackTrace();
	return null;
      }

      localSession =
	new SessionAddress(localSessionAddr,
			   localSessionPort,
			   localSessionAddr,
			   localSessionPort + 1);
	
      System.out.println("Local Session Address = " +
			 localSessionAddr);
      System.out.println("Local Session port = " +
			 localSessionPort);
    }

    // Set source description items used in RTCP SDES reports
    SourceDescription [] sdes = setSourceDescription(newMgr);

    SessionAddress localAddr = new SessionAddress();

    // Generate SSRC
    long ssrc = newMgr.generateSSRC();

    // call initSession() and startSession() of the RTPsessionManager
    int result = -1;
    try {
      // Initialize session
      newMgr.initSession(localAddr, // local address
			 ssrc,	    // default SSRC
			 sdes,	    // default user description
			 0.05,      // rtcp_bw_fraction
			 0.25);     // rtcp_sender_bw_fraction

      // Start the session, causing RTCP reports to be generated
      if (isSender) 
	newMgr.startSession(remoteSession, ttl, null);
      else
	newMgr.startSession(localSession, ttl, null);	
    } catch (SessionManagerException e){
      System.err.println("RTPSM Exception " + e.getMessage());
      e.printStackTrace();
      return null;
    } 
    catch (NumberFormatException e){
      System.err.println("NumberFormat Exception " + e.getMessage());
      e.printStackTrace();
      return null;
    } 
    catch (IOException e){
      System.err.println("IO Exception " + e.getMessage());
      e.printStackTrace();
      return null;
    } 
	
    return newMgr;
  }

  // Set source description items used in RTCP SDES reports 
  private SourceDescription[] setSourceDescription(SessionManager newMgr) {
    // Ask RTPSM to generate the local participants CNAME
    String cname = newMgr.generateCNAME();
    
    SourceDescription[] sd = new SourceDescription[4];
    for (int i = 0; i < sd.length; i++) {
      if (i == 0) {
	sd[i] = new
	  SourceDescription(SourceDescription.SOURCE_DESC_EMAIL,
			    "xiaotaow@cs.columbia.edu",
			    1,
			    false);
      }

      else if (i == 1) {
	sd[i] = new
	  SourceDescription(SourceDescription.SOURCE_DESC_NAME,
			    "xiaotaow",
			    1,
			    false);
      }

      else if (i == 2) {
	sd[i] = new 
	  SourceDescription(SourceDescription.SOURCE_DESC_CNAME,
			    cname,
			    1,
			    false);
      }
      
    }
    
    return sd;
  }

  // close SessionManager due to error 
  private void closeManager() 
  {
    if (streamSend != null) {
      streamSend.close();
      streamSend = null;
    }
	
    if (sendingMgr != null) {
      sendingMgr.closeSession("RTP Session Terminated");
      sendingMgr = null;
    }
    if (receivingMgr != null) {
      receivingMgr.closeSession("RTP Session Terminated");
      receivingMgr = null;
    }
    
  }

  // Transmit RTP data over the network
  public synchronized void transmit() {
    String errorMessage;
        
    // Create an RTP session to transmit the output of the
    // processor to the specified IP address and port no.
    errorMessage = createSendStream();
    if (errorMessage != null) {
      System.err.println(errorMessage);
      destroyProcessor();
      closeManager();
      System.exit(1);
    }
        
    System.out.println("Starting the processor");
    processor.start();
    started = true; // set flag indicating Processor has started
    displayFrame(); // display GUI for source description	
  }

  private String createSendStream() {
    // Create SessionManager to transmit RTP data 
    sendingMgr = createManager(remoteSessionAddress, remoteSessionPort, true);
    if (sendingMgr == null) 
      return "Failed to create SessionManager";
	
    try {
      // Create a sending stream within the RTP session
      streamSend = sendingMgr.createSendStream(processorOutput, 0); 
      // Start RTP data transmission over the network
      streamSend.start();
    } catch (Exception e) {
      e.printStackTrace();
      return "Exception in SendStream";
    } 
	
    return null;
  }

  // Called for REINVITE
  // Stop transmitting to the old destination and start sending to the 
  //	new destination
  public void handoff() 
  {
    try {
      InetAddress remoteSessionAddr =
	InetAddress.getByName(remoteSessionAddress);
     
      if (sendingMgr == null) {
	System.out.println("Session Manager is null");
	System.exit(1);
      }

    
      remoteSession =
	new SessionAddress(remoteSessionAddr,
			   remoteSessionPort,
			   remoteSessionAddr,
			   remoteSessionPort + 1);
      
      System.out.println("Handoff to " +
			 remoteSessionAddr.getHostName() +
			 ":" + remoteSessionPort);
	    
      // Stop transmitting RTP data to the old destination
      sendingMgr.closeSession("RTP Session Terminated");
      sendingMgr = null;

      // Start sending RTP data to the new destination
      sendingMgr = createManager(remoteSessionAddress, remoteSessionPort,
				 true);
      streamSend = sendingMgr.createSendStream(processorOutput, 0); 
      streamSend.start();
    }
    catch (Exception e) {
      System.err.println("Exception " + e.getMessage());
      e.printStackTrace();
      System.exit(-1);
    }
  }

  // Call back in the RemoteListener to notify lisener of all RTP Remote
  //	Events	(including SenderReport) 
  public void update(RemoteEvent event) 
  {
    // Inform the RTP listener (this RTP Agent) that an RTCP sender report
    //	(SR) has been received
    if (event instanceof SenderReportEvent) {
      // Get the RTCP sender report (SR)
      SenderReport sr = ((SenderReportEvent)event).getReport();

      // Get the Participant that sent this sender report
      Participant p = sr.getParticipant();
 
      if (p != null && textArea != null) {
	if (!participantList.contains(p)) {
	  System.out.println("SenderReport from " + p.getCNAME());
	  participantList.addElement(p);
	  // Display the source description on the GUI
	  textArea.append("CNAME: " + (sr.getParticipant()).getCNAME() + 
	  	  "\n");
	  Vector sdes = (sr.getParticipant()).getSourceDescription();
	  for (int i = 0; i < sdes.size(); i++) {
	  	SourceDescription des = (SourceDescription)sdes.elementAt(i);
	  	if (des.getType() == SourceDescription.SOURCE_DESC_NAME) 
	  	  textArea.append("Name: " + des.getDescription() + "\n");
	  	else if (des.getType() == SourceDescription.SOURCE_DESC_EMAIL) 
	  	  textArea.append("Email: " + des.getDescription() + "\n");	
	  }	
	  textArea.append("*****************************************************\n\n");
		
	}
      }
    }
  }

  // Call back in the RTPSessionListener to notify listener of all 
  //	ReceiveStream Events including NewReceiveStreamEvent and 
  //	ByeEvent
  public void update(ReceiveStreamEvent event){
    Player newPlayer = null;

    // Create a new player if a new recvstream is detected
    if (event instanceof NewReceiveStreamEvent){
      try{
	ReceiveStream stream = 
	  ((NewReceiveStreamEvent)event).getReceiveStream();	
	DataSource source = stream.getDataSource();
	newPlayer = Manager.createPlayer(source);
	
      } catch (Exception e){
	System.err.println("NewReceiveStreamEvent Exception " + 
			   e.getMessage());
	e.printStackTrace();
	System.exit(-1);
      }
	
      if (newPlayer == null){
	System.err.println("Can't create a Player for ReceiveStream");
	System.exit(-1);
      }
      newPlayer.start();
      System.out.println("Player started");

      playerList.addElement(newPlayer);
    }
    // Inform the RTP listener (this RTP Agent) that an RTCP BYE indication
    //	has been received.
    else if (event instanceof ByeEvent) {
      Participant p = ((ByeEvent)event).getParticipant();
      if (p != null && textArea != null) {
	System.out.println("Bye Event from " + p.getCNAME());
	
	if (participantList.contains(p)) {
	  int index = participantList.indexOf(p);
	  participantList.removeElementAt(index);

	  // Update the GUI for the source description  
	  textArea = new JTextArea(250, 650);
	  panel.add(textArea);
	  for (int i = 0; i < participantList.size(); i++) {
	    Participant pt = (Participant)participantList.elementAt(i);
	    textArea.append(" CNAME: " + pt.getCNAME() + 
			    "\n");
	    Vector sdes = pt.getSourceDescription();
	    for (int j = 0; j < sdes.size(); j++) {
	      SourceDescription des = (SourceDescription)sdes.elementAt(j);
	      if (des.getType() == SourceDescription.SOURCE_DESC_NAME) 
		textArea.append(" Name: " + des.getDescription() + "\n");
	      else if (des.getType() == SourceDescription.SOURCE_DESC_EMAIL) 
		textArea.append(" Email: " + des.getDescription() + "\n");
	    }	
	    textArea.append("*********************************************************\n\n");
	  }
	}
      }
    }
  } 

  private Object lock = new Object();

  // flag that indicates whether it is OK for Processor to transit to the 
  //	next state
  private boolean passed = true; 


  // Make the Processor wait until it is OK to transit to the next state
  private boolean handleStateTransition(Processor p, int state) {
    synchronized (lock) {
      try {
	// Wait until we get an event that confirms the
	// success of the method, or a failure event.
	while (p.getState() < state && passed) 
	  lock.wait();
      }
      catch (InterruptedException ie) {
	return false;
      }
    }
    return passed;
  }

  // Called when an event is generated by Processor
  public void controllerUpdate(ControllerEvent event) {
    // If it is OK to transit to the next Processor state
    if (event instanceof ConfigureCompleteEvent ||
	event instanceof RealizeCompleteEvent ||
	event instanceof PrefetchCompleteEvent) {
      synchronized (lock) {
	passed = true;
	lock.notifyAll();
      }
    }
    else if (event instanceof ResourceUnavailableEvent) {
      synchronized (lock) {
	passed = false;
	lock.notifyAll();
      }
    }
    // If there was an error during configure or
    // realize, the processor will be closed
    else if (event instanceof ControllerClosedEvent) {
      synchronized (lock) {
	passed = false;
	lock.notifyAll();
      }
    }	  
  }

  // Display GUI for Source Description
  private void displayFrame() {
    setTitle("Source Description");
    setSize(300, 500);
    setVisible(true);
    panel = new JPanel(new BorderLayout());
    getContentPane().add(panel, "Center");
    textArea = new JTextArea(250, 450);
    panel.add(textArea);
		
  }			

}













